@@ -209,7 +209,16 @@ generator_funs$clone_method <- function(deep = FALSE) {
 
     # Copy the old objects, fix up method environments, and put them into the
     # new binding environment.
-    binding_copies <- as.list.environment(old_slice$binding, all.names = TRUE)
+
+    # Separate active and non-active bindings. We'll copy over just the
+    # non-active bindings now; the active bindings need to be copied over with
+    # a different method later.
+    binding_names <- ls(old_slice$binding, all.names = TRUE)
+    if (!is.null(old_slice$enclosing$`.__active__`)) {
+      binding_names <- setdiff(binding_names, names(old_slice$enclosing$`.__active__`))
+    }
+
+    binding_copies <- mget(binding_names, envir = old_slice$binding)
 
     # Don't copy self, private, super, or .__enclos_env__. Note that using
     # %in% is significantly faster than setdiff() here.
@@ -218,16 +227,6 @@ generator_funs$clone_method <- function(deep = FALSE) {
 
     binding_copies <- remap_func_envs(binding_copies, old_new_enclosing_pairs)
 
-    # Separate active and non-active bindings
-    active_idx <- vapply(
-      names(binding_copies),
-      bindingIsActive,
-      env = old_slice$binding,
-      TRUE
-    )
-    active_copies  <- binding_copies[active_idx]
-    binding_copies <- binding_copies[!active_idx]
-
     if (deep) {
       binding_copies <- mapply(
         deep_clone,
@@ -237,13 +236,18 @@ generator_funs$clone_method <- function(deep = FALSE) {
       )
     }
 
-    # Copy in public and active bindings
+    # Copy in public bindings
     list2env2(binding_copies, new_slice$binding)
 
-    if (length(active_copies) > 0) {
+
+    # Now copy over active bindings, if present
+    if (!is.null(old_slice$enclosing$`.__active__`)) {
+      active_copies <- remap_func_envs(old_slice$enclosing$`.__active__`, old_new_enclosing_pairs)
+
       for (name in names(active_copies)) {
         makeActiveBinding(name, active_copies[[name]], new_slice$binding)
       }
+      new_slice$enclosing$`.__active__` <- active_copies
     }
 
     # Copy private members
